from __future__ import annotations

import contextlib
import csv
import enum
import io
import os
import pathlib
from dataclasses import dataclass, field
from datetime import datetime
from typing import Callable, Generator, Iterable, Optional, TYPE_CHECKING

from databudgie.adapter.base import QueryResult
from databudgie.compression import Compressor
from databudgie.config import BackupConfig, RestoreConfig
from databudgie.manifest.manager import Manifest
from databudgie.output import Console, default_console, Table
from databudgie.s3 import is_s3_path, optional_s3_resource, S3Location
from databudgie.utils import join_paths

if TYPE_CHECKING:
    from mypy_boto3_s3.service_resource import Bucket, S3ServiceResource

DATETIME_FORMAT = r"%Y-%m-%dT%H:%M:%S"


@enum.unique
class FileTypes(enum.Enum):
    ddl = "ddl"
    sequences = "sequences"
    data = "data"
    manifest = "manifest"

    @property
    def extension(self) -> str:
        mapping = {
            self.ddl: "sql",
            self.sequences: "json",
            self.data: "csv",
            self.manifest: "json",
        }
        return mapping[self]  # type: ignore


@dataclass
class LocalStorage:
    def write_buffer(self, path: str, buffer: io.BytesIO):
        npath = pathlib.PurePath(path)

        parent = npath.parent
        os.makedirs(parent, exist_ok=True)
        with open(path, "wb") as f:
            f.write(buffer.getbuffer())

    def path_exists(self, path: str):
        matching_objects = [dir_entry for dir_entry in os.scandir(path) if dir_entry.is_file()]
        return len(matching_objects) >= 1

    def get_file_content(
        self, path: str, selection_strategy: SelectionStrategy, file_type: FileTypes, compression: str | None = None
    ) -> FileObject | None:
        try:
            files = os.scandir(path)
        except FileNotFoundError:
            return None

        object_generator = (
            FileStat.from_stat(os.path.sep.join([path, dir_entry.name]), dir_entry.stat())
            for dir_entry in files
            if dir_entry.is_file()
        )

        target_object = selection_strategy(object_generator, file_type, compression)
        if not target_object:
            return None

        buffer = io.BytesIO()
        with open(target_object.path, "rb") as f:
            buffer.write(f.read())
        buffer.seek(0)

        return FileObject(path=str(target_object.path), content=buffer)


@dataclass
class S3Storage:
    resource: S3ServiceResource

    @classmethod
    def from_config(cls, config: BackupConfig | RestoreConfig) -> S3Storage | None:
        resource = optional_s3_resource(config)
        if not resource:
            return None

        return cls(resource=resource)

    def write_buffer(self, path: str, buffer: io.BytesIO):
        s3_location = S3Location(path)
        s3_bucket: Bucket = self.resource.Bucket(s3_location.bucket)
        s3_bucket.put_object(Key=s3_location.key, Body=buffer)

    def path_exists(self, path: str):
        s3_location = S3Location(path)
        s3_bucket: Bucket = self.resource.Bucket(s3_location.bucket)
        matching_objects = list(s3_bucket.objects.filter(Prefix=s3_location.key).all())
        return len(matching_objects) >= 1

    def get_file_content(
        self, path: str, selection_strategy: SelectionStrategy, file_type: FileTypes, compression: str | None = None
    ) -> FileObject | None:
        buffer = io.BytesIO()

        # this path.key should be a folder
        s3_location = S3Location(path)
        s3_bucket: Bucket = self.resource.Bucket(s3_location.bucket)

        object_generator = self._generate_s3_object_summaries(s3_bucket, s3_location)
        target_object = selection_strategy(object_generator, file_type, compression)
        if not target_object:
            return None

        s3_bucket.download_fileobj(str(target_object.path), buffer)
        path = f"s3://{s3_location.bucket}/{target_object.path}"

        buffer.seek(0)
        return FileObject(path, buffer)

    def _generate_s3_object_summaries(self, s3_bucket, s3_location):
        for o in s3_bucket.objects.filter(Prefix=s3_location.key):
            object_summary = FileStat(o.key, o.last_modified)
            if str(object_summary.path.parent) == s3_location.key:
                yield object_summary


@dataclass
class StorageBackend:
    local_storage: LocalStorage
    s3_storage: S3Storage | None = None

    timestamp: datetime = field(default_factory=lambda: datetime.utcnow())

    manifest: Manifest | None = None
    events: dict[str, TableInfo] = field(default_factory=dict)

    record_stats: bool = False
    perform_writes: bool = False

    @classmethod
    def from_config(
        cls,
        config: BackupConfig | RestoreConfig,
        manifest: Manifest | None = None,
        perform_writes=True,
        record_stats=False,
    ) -> StorageBackend:
        return cls(
            local_storage=LocalStorage(),
            s3_storage=S3Storage.from_config(config),
            manifest=manifest,
            perform_writes=perform_writes,
            record_stats=record_stats,
        )

    def check_manifest(self, table_name: str):
        return self.manifest and table_name in self.manifest

    def choose_storage(self, path: str) -> LocalStorage | S3Storage:
        if is_s3_path(path):
            assert self.s3_storage
            return self.s3_storage
        if self.local_storage:
            return self.local_storage

        raise ValueError("No storage backend found")

    def write_buffer(
        self,
        path: str,
        _buffer: io.BytesIO | QueryResult,
        *,
        file_type: FileTypes,
        name: str | None = None,
        compression: str | None = None,
    ):
        if isinstance(_buffer, QueryResult):
            buffer = _buffer.buffer
        else:
            buffer = _buffer

        if name and self.record_stats:
            table_info = self.events.setdefault(name, TableInfo(name=name))
            table_info.note_file_type(file_type)

            if file_type == FileTypes.data:
                assert isinstance(_buffer, QueryResult)
                row_count = _buffer.row_count
                table_info.rows = row_count

        filename = join_paths(path, generate_filename(self.timestamp, file_type=file_type, compression=compression))
        if self.perform_writes:
            final_buffer = Compressor.get_with_name(compression).compress(buffer)

            storage = self.choose_storage(filename)

            storage.write_buffer(filename, final_buffer)

            # `name` is primarily omitted for things spanning individual tables, like schemas.
            if name and self.manifest and file_type == FileTypes.data:
                self.manifest.record(name, filename)

        return filename

    def path_exists(self, path: str) -> bool:
        storage = self.choose_storage(path)
        return storage.path_exists(path)

    @contextlib.contextmanager
    def get_file_content(
        self,
        path: str,
        strategy: str,
        *,
        name: str | None = None,
        file_type: FileTypes,
        compression=None,
    ) -> Generator[FileObject | None, None, None]:
        storage = self.choose_storage(path)
        selection_strategy = FileSelectionStrategy.by_name(strategy)

        file_object = storage.get_file_content(path, selection_strategy, file_type=file_type, compression=compression)
        if not file_object:
            yield None
            return

        cbuffer = Compressor.get_with_name(compression).extract(file_object.content)

        yield FileObject(path=path, content=cbuffer)

        if name and self.record_stats:
            table_info = self.events.setdefault(name, TableInfo(name=name))
            table_info.note_file_type(file_type)

            if file_type == FileTypes.data:
                cbuffer.seek(0)
                row_count = len(list(csv.DictReader(io.TextIOWrapper(cbuffer))))
                table_info.rows = row_count

        if self.manifest and name and self.perform_writes and file_type == file_type.data:
            self.manifest.record(name, file_object.path)
        cbuffer.close()

    def print_stats(self, console: Console = default_console):
        table = Table(title="Stats")

        table.add_column("Table", justify="right", style="cyan", no_wrap=True)
        table.add_column("DDL", justify="right", style="green")
        table.add_column("Sequences", justify="right", style="green")
        table.add_column("Data", justify="right", style="green")
        table.add_column("Rows", justify="right")

        for row in self.events.values():
            table.add_row(
                row.name,
                "✓" if row.ddl else "",
                "✓" if row.sequences else "",
                "✓" if row.data else "",
                str(row.rows) if row.rows is not None else "",
            )

        console.print(table)


@dataclass
class TableInfo:
    name: str
    ddl: bool = False
    sequences: bool = False
    data: bool = False
    rows: int | None = None

    def note_file_type(self, file_type: FileTypes):
        if file_type == file_type.ddl:
            self.ddl = True
        elif file_type == file_type.sequences:
            self.sequences = True
        elif file_type == file_type.data:
            self.data = True


@dataclass(frozen=True)
class FileObject:
    path: str
    content: io.BytesIO


@dataclass(frozen=True)
class FileStat:
    path_str: str
    last_modified: datetime

    @property
    def path(self):
        return pathlib.PurePath(self.path_str)

    @classmethod
    def from_stat(cls, name, stat: os.stat_result):
        return cls(name, datetime.fromtimestamp(stat.st_mtime))


SelectionStrategy = Callable[[Iterable[FileStat], FileTypes, Optional[str]], Optional[FileStat]]


class FileSelectionStrategy:
    @classmethod
    def by_name(cls, name: str) -> SelectionStrategy:
        valid_strategies = {
            "use_latest_metadata": cls.use_metadata_strategy,
            "use_latest_filename": cls.use_filename_strategy,
        }
        return valid_strategies[name]

    @staticmethod
    def use_filename_strategy(
        available_objects: Iterable[FileStat], file_type: FileTypes, compression=None
    ) -> FileStat | None:
        def restore_filename(timestamp, *, file_type: FileTypes, compression=None):
            extension = file_type.extension
            filetype = Compressor.get_with_name(compression).compose_filetype(extension)
            return datetime.strptime(timestamp, f"{DATETIME_FORMAT}.{filetype}")

        objects_by_filename = {obj.path.name: obj for obj in available_objects}
        ordered_filenames = sorted(
            objects_by_filename.keys(),
            key=lambda x: restore_filename(x, file_type=file_type, compression=compression),
            reverse=True,
        )

        if ordered_filenames:
            return objects_by_filename[ordered_filenames[0]]
        return None

    @staticmethod
    def use_metadata_strategy(
        available_objects: Iterable[FileStat], file_type: FileTypes, compression=None
    ) -> FileStat | None:
        objects_by_last_modified_date = {s3_object.last_modified: s3_object for s3_object in available_objects}
        ordered_last_modified_dates = sorted(objects_by_last_modified_date.keys(), reverse=True)

        if ordered_last_modified_dates:
            return objects_by_last_modified_date[ordered_last_modified_dates[0]]
        return None


def generate_filename(timestamp, *, file_type: FileTypes, compression=None):
    extension = file_type.extension
    filetype = Compressor.get_with_name(compression).compose_filetype(extension)
    return timestamp.strftime(f"{DATETIME_FORMAT}.{filetype}")



from __future__ import annotations

import contextlib
import csv
import io
import warnings
from dataclasses import dataclass, field, replace
from typing import Any, cast, Generator, Sequence

import sqlalchemy
from sqlalchemy import inspect, MetaData, Table, text
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session

from databudgie.output import Console, default_console
from databudgie.table_op import TableOp
from databudgie.utils import join_paths, parse_table, wrap_buffer


@dataclass
class Adapter:
    """Root class designating a shared interface for operating on different databases.

    Any database-agnostic functionality should live here, while more optimized
    database-specific implementations should be implemented on specialized `Adaptor`
    subclasses.

    Uses native Python CSV methods with a lightweight/naive type conversion on insert.
    """

    session: Session

    @classmethod
    def get_adapter(cls, session: Session, dialect: str | None = None) -> Adapter:
        """Determine an interface based on the dialect name from the Session (or an explicit string).

        Examples:
            >>> from databudgie.adapter.postgres import PostgresAdapter
            >>> adapter = Adapter.get_adapter(None, "postgres")
            >>> isinstance(adapter, PostgresAdapter)
            True

            >>> from databudgie.adapter import Adapter
            >>> adapter = Adapter.get_adapter(None, "python")
            >>> isinstance(adapter, Adapter)
            True
        """
        if dialect is None:
            dialect = session.get_bind().dialect.name

        if dialect in ("postgres", "postgresql"):
            from databudgie.adapter.postgres import PostgresAdapter

            return PostgresAdapter(session)

        return cls(session)

    def export_query(self, query: str) -> QueryResult:
        def query_database(session: Session, query: str) -> Generator[Sequence[Any], None, None]:
            cursor = session.execute(text(query))

            columns: Sequence[str] = list(cursor.keys())
            yield columns

            yield from cursor

        result = QueryResult()
        with result.text_buffer() as text_buffer:
            writer = csv.writer(text_buffer, quoting=csv.QUOTE_MINIMAL)

            i = 0
            for i, row in enumerate(query_database(self.session, query), start=1):
                writer.writerow(row)

        result.row_count = i
        return result

    def import_csv(self, csv_file: io.TextIOBase, table: str):
        reader = csv.DictReader(csv_file, quoting=csv.QUOTE_MINIMAL)

        prepared_rows: list[dict] = []
        for row in reader:
            new_row: dict[str, Any] = dict(row)
            for key, value in new_row.items():
                if value.lower() == "true":
                    new_row[key] = True
                elif value.lower() == "false":
                    new_row[key] = False
                elif value == "":
                    new_row[key] = None

            prepared_rows.append(new_row)

        schema, table_name = parse_table(table)

        engine = cast(Engine, self.session.get_bind())
        metadata = MetaData()
        metadata.reflect(engine, schema=schema)
        table_ref = Table(table_name, metadata, autoload=True, autoload_with=engine, schema=schema)

        with engine.connect() as conn:
            conn.execute(table_ref.insert(), prepared_rows)
            conn.execute(text("commit"))

    def export_schema_ddl(self, name: str):
        raise NotImplementedError()  # pragma: no cover

    def export_table_ddl(self, table_name: str, console: Console = default_console):
        raise NotImplementedError()  # pragma: no cover

    def truncate_table(self, table: str):
        try:
            self.session.execute(text(f"TRUNCATE TABLE {table} CASCADE"))
            self.session.commit()
        except sqlalchemy.exc.ProgrammingError:
            self.session.rollback()

    def reset_database(self):
        """Reset the database in a database-backend agnostic way.

        Create a temp database, connect to it, drop the target database, and drop
        the temp database.

        This method suffers from being rather prone to failure, but is better
        than nothing!
        """
        raise NotImplementedError()

    def collect_existing_tables(self) -> list[str]:
        """Find the set of all user-defined tables in a database."""
        connection = self.session.connection()

        metadata = MetaData()
        insp = inspect(connection)
        for schema in insp.get_schema_names():
            # Seems to be a generally cross-database compatible filter.
            if schema in ("information_schema", "pg_catalog"):
                continue

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                metadata.reflect(bind=connection, schema=schema)

        return [table.fullname for table in metadata.sorted_tables]

    def collect_table_dependencies(self, table_op: TableOp, console: Console = default_console) -> list[str]:
        raise NotImplementedError()

    def collect_table_sequences(self) -> dict[str, list[str]]:
        raise NotImplementedError()

    def collect_sequence_value(self, sequence_name: str) -> int:
        raise NotImplementedError()

    def restore_sequence_value(self, sequence_name: str, value: int) -> int:
        raise NotImplementedError()

    def materialize_table_dependencies(
        self,
        table_ops: list[TableOp],
        reverse: bool = False,
        console: Console = default_console,
    ) -> list[TableOp]:
        tables = set()
        dependent_table_ops = []
        for table_op in table_ops:
            tables.add(table_op.full_name)

            if not table_op.raw_conf.follow_foreign_keys:
                continue

            dependent_tables = self.collect_table_dependencies(table_op=table_op, console=console)
            for dependent_table in dependent_tables:
                if dependent_table not in tables:
                    table_location = join_paths(table_op.location(), "{table}")
                    conf = replace(table_op.raw_conf, name=dependent_table, location=table_location)

                    dependent_table_op = TableOp.from_name(dependent_table, conf)
                    tables.add(dependent_table)
                    dependent_table_ops.append(dependent_table_op)

        if reverse:
            # The original `table_ops` list comes to us already reverse
            # ordered, so we need to preserve it's original order, and
            # just invert the net-new tables and put them first.
            return list(reversed(dependent_table_ops)) + table_ops

        return table_ops + dependent_table_ops


@dataclass
class QueryResult:
    buffer: io.BytesIO = field(default_factory=io.BytesIO)
    row_count: int = 0

    @contextlib.contextmanager
    def binary_buffer(self):
        yield self.buffer
        self.buffer.seek(0)

    @contextlib.contextmanager
    def text_buffer(self):
        with wrap_buffer(self.buffer) as text_buffer:
            yield text_buffer



from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mypy_boto3_s3 import S3ServiceResource

    from databudgie.config import BackupConfig, RestoreConfig, S3Config


def optional_s3_resource(config: BackupConfig | RestoreConfig) -> S3ServiceResource | None:
    if config_uses_s3(config):
        return s3_resource(config.s3)
    return None


def s3_resource(config: S3Config | None = None) -> S3ServiceResource:
    try:
        import boto3
    except ImportError:
        raise RuntimeError('Use of S3 requires the "s3" python extra')

    # Boto loads all config as environment variables by default, this config
    # section can be entirely optional.
    if not config:
        session = boto3.session.Session()
    else:
        session = boto3.session.Session(
            aws_access_key_id=config.aws_access_key_id,
            aws_secret_access_key=config.aws_secret_access_key,
            profile_name=config.profile,
            region_name=config.region,
        )

    s3: S3ServiceResource = session.resource("s3")
    return s3


def config_uses_s3(config: BackupConfig | RestoreConfig):
    if config.root_location and is_s3_path(config.root_location):
        return True

    for table_config in config.tables:
        if is_s3_path(table_config.location):
            return True
    return False


def is_s3_path(path: str):
    return path.startswith("s3://")


class S3Location:
    """Easily parse an S3 URL into Bucket and Key.

    Example:
        >>> loc = S3Location("s3://my-s3-bucket/raw_upload/sample.csv")
        >>> loc.bucket
        'my-s3-bucket'
        >>> loc.key
        'raw_upload/sample.csv'
    """

    def __init__(self, url: str):
        self._parsed = urllib.parse.urlparse(url)

    @property
    def bucket(self):
        return self._parsed.netloc

    @property
    def key(self):
        return self._parsed.path.lstrip("/")



from __future__ import annotations

import io
import json
from typing import Sequence

from sqlalchemy.orm import Session

from databudgie.adapter import Adapter
from databudgie.config import BackupConfig, BackupTableConfig
from databudgie.output import Console, default_console, Progress
from databudgie.storage import FileTypes, StorageBackend
from databudgie.table_op import expand_table_ops, TableOp
from databudgie.utils import capture_failures, join_paths


def backup_all(
    session: Session,
    backup_config: BackupConfig,
    storage: StorageBackend | None = None,
    console: Console = default_console,
):
    """Perform backup on all tables in the config.

    Arguments:
        session: A SQLAlchemy session with the PostgreSQL database from which to query data.
        backup_config: config object mapping table names to their query and location.
        storage: Storage backend to use for backing up the data.
        console: Console used for output
    """
    if storage is None:
        storage = StorageBackend.from_config(backup_config)
    adapter = Adapter.get_adapter(session, backup_config.adapter)

    existing_tables = adapter.collect_existing_tables()
    table_ops = expand_table_ops(
        session,
        backup_config.tables,
        existing_tables,
        storage=storage,
        console=console,
        warn_for_unused_tables=True,
    )

    table_ops = adapter.materialize_table_dependencies(
        table_ops,
        console=console,
    )

    backup_ddl(
        backup_config,
        table_ops,
        adapter=adapter,
        storage=storage,
        console=console,
    )
    backup_sequences(
        table_ops,
        adapter=adapter,
        storage=storage,
        console=console,
    )
    backup_tables(
        table_ops=table_ops,
        adapter=adapter,
        storage=storage,
        console=console,
    )


def backup_ddl(
    backup_config: BackupConfig,
    table_ops: list[TableOp[BackupTableConfig]],
    *,
    adapter: Adapter,
    storage: StorageBackend,
    console: Console = default_console,
):
    if not backup_config.ddl.enabled:
        return

    # Backup schemas first
    schema_names = set()
    schemas = []

    for table_op in table_ops:
        schema_op = table_op.schema_op()
        if schema_op.name in schemas:
            continue

        if not table_op.raw_conf.ddl:
            continue

        schema_names.add(schema_op.name)
        schemas.append(schema_op)

    with Progress(console) as progress:
        table_names = []

        total = len(schemas) + len(table_ops)
        task = progress.add_task("Backing up schema DDL", total=total)

        for schema_op in schemas:
            progress.update(task, description=f"Backing up schema DDL: {schema_op.name}")

            buffer = io.BytesIO(adapter.export_schema_ddl(schema_op.name))
            filename = storage.write_buffer(
                join_paths(backup_config.ddl.location, schema_op.location()),
                buffer,
                file_type=FileTypes.ddl,
            )

            console.trace(f"Wrote {schema_op.name} to {filename}")

        console.info("Finished backing up schema DDL")

        for table_op in table_ops:
            if not table_op.raw_conf.ddl:
                continue

            progress.update(task, description=f"Backing up DDL: {table_op.full_name}")
            result = adapter.export_table_ddl(table_op.full_name)

            filename = storage.write_buffer(
                join_paths(backup_config.ddl.location, table_op.location()),
                io.BytesIO(result),
                file_type=FileTypes.ddl,
                name=table_op.full_name,
            )

            console.trace(f"Uploaded {table_op.full_name} to {filename}")
            table_names.append(table_op.full_name)

    console.info("Finished backing up DDL")

    # On the restore-side, the tables may not already exist (at the extreme, you
    # might start with an empty database), so we need to record the set of tables
    # being backed up.
    manifest_data = json.dumps(table_names).encode("utf-8")
    with io.BytesIO(manifest_data) as buffer:
        storage.write_buffer(backup_config.ddl.location, buffer, file_type=FileTypes.manifest)


def backup_sequences(
    table_ops: list[TableOp[BackupTableConfig]],
    *,
    adapter: Adapter,
    storage: StorageBackend,
    console: Console = default_console,
):
    has_sequences = any(o.raw_conf.sequences for o in table_ops)
    if not has_sequences:
        return

    table_sequences = adapter.collect_table_sequences()

    with Progress(console) as progress:
        task = progress.add_task("Backing up sequence positions", total=len(table_ops))

        for table_op in table_ops:
            progress.update(task, description=f"Backing up sequence position: {table_op.full_name}")

            if not table_op.raw_conf.sequences:
                continue

            sequences = table_sequences.get(table_op.full_name)
            if not sequences:
                continue

            sequence_values = {}
            for sequence in sequences:
                sequence_values[sequence] = adapter.collect_sequence_value(sequence)

            result = json.dumps(sequence_values).encode("utf-8")

            path = table_op.location()

            filename = storage.write_buffer(
                join_paths(path, "sequences"),
                io.BytesIO(result),
                file_type=FileTypes.sequences,
                name=table_op.full_name,
            )

            console.trace(f"Wrote {table_op.full_name} sequences to {filename}")

    console.info("Finished backing up sequence positions")


def backup_tables(
    table_ops: Sequence[TableOp],
    storage: StorageBackend,
    *,
    adapter: Adapter,
    console: Console = default_console,
) -> None:
    with Progress(console) as progress:
        task = progress.add_task("Backing up tables", total=len(table_ops))

        for table_op in table_ops:
            progress.update(task, description=f"Backing up table: {table_op.full_name}")

            if not table_op.raw_conf.data:
                continue

            with capture_failures(strict=table_op.raw_conf.strict):
                backup(
                    table_op=table_op,
                    storage=storage,
                    adapter=adapter,
                    console=console,
                )

    console.info("Finished Backing up tables")


def backup(
    *,
    table_op: TableOp[BackupTableConfig],
    adapter: Adapter,
    storage: StorageBackend,
    console: Console = default_console,
):
    """Dump query contents to S3 as a CSV file.

    Arguments:
        config: The raw backup configuration.
        table_op: The table operation being acted up on.
        adapter: the selected behavior adapter
        storage: the storage backend to use for backing up the data.
        console: Console used for output
    """
    buffer = adapter.export_query(table_op.query())

    # path.join will handle optionally trailing slashes in the location
    compression = table_op.raw_conf.compression

    filename = storage.write_buffer(
        table_op.location(), buffer, file_type=FileTypes.data, name=table_op.full_name, compression=compression
    )

    console.trace(f"Uploaded {table_op.full_name} to {filename}")



from datetime import datetime

import sqlalchemy


def create_manifest_table(metadata, tablename, schema=None):
    return sqlalchemy.Table(
        tablename,
        metadata,
        sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True, autoincrement=True),
        sqlalchemy.Column("transaction", sqlalchemy.types.Integer, nullable=False, index=True),
        sqlalchemy.Column("action", sqlalchemy.types.String, nullable=False, index=True),  # backup or restore
        sqlalchemy.Column("table", sqlalchemy.types.String, nullable=False, index=True),
        sqlalchemy.Column("file_path", sqlalchemy.types.String, nullable=False),
        sqlalchemy.Column(
            "timestamp",
            sqlalchemy.types.DateTime(timezone=True),
            default=datetime.utcnow,
            server_default=sqlalchemy.text("CURRENT_TIMESTAMP"),
            nullable=False,
        ),
        schema=schema,
    )


class DatabudgieManifestMixin:
    id = sqlalchemy.Column(sqlalchemy.Integer, primary_key=True, autoincrement=True)
    transaction = sqlalchemy.Column(sqlalchemy.types.Integer, nullable=False, index=True)
    action = sqlalchemy.Column(sqlalchemy.types.String, nullable=False, index=True)  # backup or restore
    table = sqlalchemy.Column(sqlalchemy.types.String, nullable=False, index=True)
    file_path = sqlalchemy.Column(sqlalchemy.types.String, nullable=False)
    timestamp = sqlalchemy.Column(
        sqlalchemy.types.DateTime(timezone=True),
        default=datetime.utcnow,
        server_default=sqlalchemy.text("CURRENT_TIMESTAMP"),
        nullable=False,
    )



import contextlib
import io
import os
import shlex
import shutil
import subprocess  # nosec
from typing import cast, Dict, List

from psycopg2._psycopg import cursor
from sqlalchemy import text
from sqlalchemy.engine import create_engine, Engine
from sqlalchemy.engine.url import URL

from databudgie.adapter.base import Adapter, QueryResult
from databudgie.output import Console, default_console
from databudgie.table_op import TableOp


def update_url(url, database=None):
    try:
        return url.set(database=database)
    except AttributeError:
        return URL(
            drivername=url.drivername,
            username=url.username,
            password=url.password,
            host=url.host,
            port=url.port,
            database=database or url.database,
        )


class PostgresAdapter(Adapter):
    def export_query(self, query: str) -> QueryResult:
        engine: Engine = cast(Engine, self.session.get_bind())

        result = QueryResult()
        with result.binary_buffer() as buffer:
            with contextlib.closing(engine.raw_connection()) as conn:
                with cast(cursor, conn.cursor()) as cursor_:
                    copy = f"COPY ({query}) TO STDOUT CSV HEADER"

                    cursor_.copy_expert(copy, buffer)
                    result.row_count = cursor_.rowcount

        return result

    def import_csv(self, csv_file: io.TextIOBase, table: str):
        engine: Engine = cast(Engine, self.session.get_bind())

        # Reading the header line from the buffer removes it for the ingest
        columns: List[str] = [f'"{c}"' for c in csv_file.readline().strip().split(",")]
        copy = "COPY {table} ({columns}) FROM STDIN CSV".format(table=table, columns=",".join(columns))

        with contextlib.closing(engine.raw_connection()) as conn:
            with cast(cursor, conn.cursor()) as cursor_:
                cursor_.copy_expert(copy, csv_file)
                conn.commit()

    def export_schema_ddl(self, name: str, console: Console = default_console) -> bytes:
        if not shutil.which("pg_dump"):
            console.warn("Could not find pg_dump, falling back to SQLAlchemy implementation.")
            return super().export_schema_ddl(name)

        url = self.session.connection().engine.url
        result = pg_dump(url, f"--schema-only --schema={name} --exclude-table={name}.*")
        return result.replace(f"CREATE SCHEMA {name};".encode(), f"CREATE SCHEMA IF NOT EXISTS {name};".encode())

    def export_table_ddl(self, table_name: str, console: Console = default_console):
        if not shutil.which("pg_dump"):
            console.warn("Could not find pg_dump, falling back to SQLAlchemy implementation.")
            return super().export_table_ddl(table_name)

        url = self.session.connection().engine.url

        return pg_dump(url, f"--schema-only -t {table_name}")

    def reset_database(self) -> None:
        """Attempt to kill the existing database and bring it back up."""
        connection = self.session.connection()

        assert connection.engine.url
        url: URL = connection.engine.url
        database = cast(str, url.database)

        # "template1" is a *special* internal postgres database that we can be guaranteed
        # to connect to after having dropped (potentially) all other available databases.
        # "template0", used below is not allowed to be connected to.
        template_url = update_url(url, database="template1")

        template_engine = create_engine(template_url)
        with template_engine.connect().execution_options(isolation_level="AUTOCOMMIT") as connection:
            kill_pids = text("SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname = :database;")
            connection.execute(kill_pids, {"database": database})

            connection.execute(text(f"DROP DATABASE {database}"))

            # "template0" is an even more special database. You cannot create databases
            # from a template while there are active connections to it.
            connection.execute(text(f"CREATE DATABASE {database} template=template0"))

        self.session.invalidate()

    def collect_existing_tables(self, console: Console = default_console) -> List[str]:
        """Find the set of all user-defined tables in a database."""
        if "FALLBACK_SQLALCHEMY_TABLE_COLLECTION" in os.environ:
            console.warn("Using SQLAlchemy to collect tables.")
            return super().collect_existing_tables()

        collect_tables = text(
            """
            with recursive fk_tree as (
                -- All tables not referencing anything else
                select t.oid      as reloid,
                    t.relname  as table_name,
                    s.nspname  as schema_name,
                    null::text COLLATE "C" as referenced_table_name,
                    null::text COLLATE "C" as referenced_schema_name,
                    1          as level
                from pg_class t
                join pg_namespace s on s.oid = t.relnamespace
                where relkind = 'r'
                and not exists(select * from pg_constraint where contype = 'f' and conrelid = t.oid)
                and s.nspname not in ('pg_catalog', 'information_schema')

                union all

                select ref.oid,
                    ref.relname,
                    rs.nspname,
                    p.table_name,
                    p.schema_name,
                    p.level + 1
                from pg_class ref
                join pg_namespace rs on rs.oid = ref.relnamespace
                join pg_constraint c on c.contype = 'f' and c.conrelid = ref.oid
                join fk_tree p on p.reloid = c.confrelid
                where ref.oid != p.reloid -- do not enter to tables referencing theirselves.
            ),
            all_tables as (
                -- this picks the highest level for each table
                select
                    schema_name,
                    table_name,
                    level,
                    row_number() over (partition by schema_name, table_name order by level desc) as last_table_row
                from fk_tree
            )
            select schema_name || '.' || table_name, level
            from all_tables at
            where last_table_row = 1
            order by level;
            """
        )

        results = self.session.execute(collect_tables)
        return [row[0] for row in results]

    def collect_table_dependencies(self, table_op: TableOp, console: Console = default_console) -> List[str]:
        """Find the set of tables dependent on the set of input tables."""
        collect_tables = text(
            """
            with recursive fk_tree as (
                 select pg_class.oid      as table_oid,
                     pg_namespace.nspname as schema_name,
                     pg_class.relname     as table_name,
                     1                    as level
                 from pg_class
                 join pg_namespace pg_namespace on pg_namespace.oid = pg_class.relnamespace
                 where pg_namespace.nspname = :schema and pg_class.relname = :table_name
                 union all
                 select pg_class.oid      as table_oid,
                     pg_namespace.nspname as schema_name,
                     pg_class.relname     as table_name,
                     fk_tree.level + 1          as level
                 from fk_tree
                 join pg_constraint on pg_constraint.contype = 'f' and pg_constraint.conrelid = fk_tree.table_oid
                 join pg_class on pg_class.oid = pg_constraint.confrelid
                 join pg_namespace on pg_namespace.oid = pg_class.relnamespace
                 where fk_tree.table_oid != pg_class.oid -- do not enter to tables referencing theirselves.
            ),
            all_tables as (
                -- this picks the **highest** level for each table
                select
                    schema_name,
                    table_name,
                    level,
                    row_number() over (partition by schema_name, table_name order by level desc) as last_table_row
                from fk_tree
            )
            select schema_name || '.' || table_name
            from all_tables at
            where last_table_row = 1
            order by level;
            """
        )

        results = self.session.execute(
            collect_tables, params={"schema": table_op.schema, "table_name": table_op.table_name}
        )

        return [row[0] for row in results]

    def collect_table_sequences(self) -> Dict[str, List[str]]:
        sequences = self.session.execute(
            text(
                """
                SELECT
                    CASE
                        WHEN seq_ns.nspname = 'public' THEN seq.relname
                        ELSE concat(seq_ns.nspname, '.', seq.relname)
                    END AS fq_sequence_name,
                    concat(tab_ns.nspname, '.', tab.relname) AS fq_table_name
                FROM pg_class seq
                JOIN pg_namespace seq_ns ON seq.relnamespace = seq_ns.oid
                JOIN pg_depend d ON d.objid = seq.oid AND d.deptype = 'a'
                JOIN pg_class tab ON d.objid = seq.oid AND d.refobjid = tab.oid
                JOIN pg_namespace tab_ns on tab.relnamespace = tab_ns.oid
                WHERE seq.relkind = 'S'
                """
            )
        )
        result: Dict = {}
        for sequence in sequences:
            result.setdefault(sequence.fq_table_name, []).append(sequence.fq_sequence_name)
        return result

    def collect_sequence_value(self, sequence_name: str) -> int:
        return cast(int, self.session.execute(text(f"SELECT last_value from {sequence_name}")).scalar())  # noqa: S608

    def restore_sequence_value(self, sequence_name: str, value: int) -> int:
        return cast(int, self.session.execute(text(f"SELECT setval('{sequence_name}', {value})")).scalar())


def pg_dump(url: URL, rest: str = "", no_comments=True) -> bytes:
    parts = [f"pg_dump -h {url.host} -p {url.port} -U {url.username} -d {url.database} --no-password"]

    if no_comments:
        parts.append("--no-comments")

    parts.append(rest)

    raw_command = " ".join(parts)

    command = shlex.split(raw_command)

    try:
        result = subprocess.run(  # nosec
            command, capture_output=True, env={**os.environ, "PGPASSWORD": str(url.password or "")}, check=True
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(e.stderr)

    return result.stdout



from __future__ import annotations

import abc
import typing
from dataclasses import dataclass, field, fields
from typing import Any

from databudgie.utils import join_paths

T = typing.TypeVar("T", "BackupTableConfig", "RestoreTableConfig")
F = typing.TypeVar("F")


class ConfigError(ValueError):
    """Raise for invalid or incomplete config."""


class ConfigStack:
    def __init__(self, *configs: dict):
        self.configs: tuple[dict, ...] = configs

    def __getitem__(self, key):
        for config in self.configs:
            if config and key in config:
                return config[key]
        raise KeyError(key)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        for config in self.configs:
            if config and key in config:
                return True
        return False

    def push(self, config: dict):
        return ConfigStack(config, *self.configs)


class Config(metaclass=abc.ABCMeta):
    def to_dict(self) -> dict:
        result = {}
        for f in fields(self):
            v = getattr(self, f.name)
            if isinstance(v, Config):
                value: Any = v.to_dict()
            elif isinstance(v, list):
                value = [v.to_dict() if isinstance(v, Config) else v for v in v]
            elif isinstance(v, dict):
                value = {k: v.to_dict() if isinstance(v, Config) else v for k, v in v.items()}
            else:
                value = v

            result[f.name] = value
        return result


def from_partial(cls: typing.Callable[..., F], **kwargs) -> F:
    """Create a new instance of cls with the given kwargs.

    This is useful for creating a new instance of a class with only some of the
    fields set.
    """
    cleaned_kwargs = {k: v for k, v in kwargs.items() if v is not None}
    return cls(**cleaned_kwargs)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Optional Configs which have default values built-in
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


@dataclass
class DDLConfig(Config):
    enabled: bool = False
    location: str = "backups/ddl"
    clean: bool = False
    strategy: str = "use_latest_filename"

    @classmethod
    def from_dict(cls, ddl_config: dict | bool | str, root_location: str | None = None):
        if isinstance(ddl_config, (bool, str)):
            expanded_ddl_config: dict[str, Any] = {"enabled": bool(ddl_config)}
        else:
            expanded_ddl_config = ddl_config

        location = compose_root_location(root_location, expanded_ddl_config.get("location"), default="backups/ddl")

        # Splat into a new dict so we can override `location` without mutating
        # the original input (which may be re-read later in config parsing)
        final_ddl_config = {**expanded_ddl_config, "location": location}

        if "clean" in final_ddl_config:
            final_ddl_config["clean"] = bool(final_ddl_config["clean"])

        return from_partial(cls, **final_ddl_config)


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Core configuration models
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


@dataclass
class RootConfig(Config):
    backup: BackupConfig
    restore: RestoreConfig

    @classmethod
    def from_dict(cls, raw_config: dict):
        config = ConfigStack(raw_config)
        return cls.from_stack(config)

    @classmethod
    def from_stack(cls, stack: ConfigStack):
        backup_config = stack.get("backup", {})
        backup = BackupConfig.from_stack(stack.push(backup_config))

        restore_config = stack.get("restore", {})
        restore = RestoreConfig.from_stack(stack.push(restore_config))
        return cls(backup=backup, restore=restore)


@dataclass
class TableParentConfig(typing.Generic[T], Config):
    tables: list[T]
    connections: dict[str, Connection]
    ddl: DDLConfig

    connection: Connection | None = None
    manifest: str | None = None

    s3: S3Config | None = None
    root_location: str | None = None
    adapter: str | None = None

    @classmethod
    @abc.abstractmethod
    def get_child_class(cls):
        pass

    @classmethod
    def from_stack(cls, stack: ConfigStack):
        connection = Connection.from_raw(stack.get("url") or stack.get("connection"), name="default")
        root_location = stack.get("root_location")

        tables_config: list = normalize_table_config(stack.get("tables", []))
        table_class = cls.get_child_class()
        tables = [table_class.from_stack(stack.push(tbl_conf), root_location) for tbl_conf in tables_config]

        # manifest defauls to None
        manifest: str | None = stack.get("manifest")

        ddl = DDLConfig.from_dict(stack.get("ddl", {}), root_location)

        # Optional integration configs
        s3 = S3Config.from_dict(stack.get("s3"))

        adapter = stack.get("adapter")

        connections = Connection.from_collection(stack.get("connections"))

        return cls(
            connection=connection,
            tables=tables,
            manifest=manifest,
            s3=s3,
            ddl=ddl,
            root_location=root_location,
            adapter=adapter,
            connections=connections,
        )


@dataclass
class Connection(Config):
    name: str
    url: str | dict

    @classmethod
    def from_raw(cls, raw: str | dict | None, *, name: str | None = None):
        if raw is None:
            return None

        if isinstance(raw, str):
            if name is None:
                raise ConfigError(f"Connection '{raw}' requires a name")
            return cls(name="default", url=raw)

        if name is None:
            raise ConfigError(f"Connection '{raw}' requires a name")

        if "url" in raw:
            url = raw["url"]
        else:
            url = {k: v for k, v in raw.items() if k != "name"}

        return cls(name=name or "default", url=url)

    @classmethod
    def from_collection(cls, collection: list | dict | None) -> dict[str, Connection]:
        if collection is None:
            return {}

        if isinstance(collection, list):
            connections = []
            names = set()
            for c in collection:
                connection = Connection.from_raw(c, name=c.get("name"))
                assert connection is not None

                if connection.name in names:
                    raise ConfigError(f"Detected more than one connection with the same name: {connection.name}")
                names.add(connection.name)
                connections.append(connection)

            return {c.name: c for c in connections}

        return {k: Connection.from_raw(c, name=k) for k, c in collection.items()}


@dataclass
class BackupTableConfig(Config):
    name: str
    location: str = "backups/{table}"
    query: str = "select * from {table}"
    compression: str | None = None
    exclude: list = field(default_factory=list)
    ddl: bool = True
    sequences: bool = True
    data: bool = True
    follow_foreign_keys: bool = False
    strict: bool = False

    @classmethod
    def from_stack(cls, stack: ConfigStack, root_location: str | None = None):
        ddl = stack.get("ddl", True)
        if isinstance(ddl, dict):
            ddl = ddl["enabled"]

        location = compose_root_location(root_location, stack.get("location"), default="backups/{table}")

        return from_partial(
            cls,
            name=stack.get("name"),
            location=location,
            query=stack.get("query"),
            compression=stack.get("compression"),
            exclude=stack.get("exclude"),
            sequences=bool(stack.get("sequences", True)),
            data=bool(stack.get("data", True)),
            ddl=bool(ddl),
            follow_foreign_keys=bool(stack.get("follow_foreign_keys", False)),
            strict=bool(stack.get("strict", False)),
        )


@dataclass
class BackupConfig(TableParentConfig[BackupTableConfig]):
    @classmethod
    def get_child_class(cls):
        return BackupTableConfig


@dataclass
class RestoreTableConfig(Config):
    name: str
    location: str = "backups/{table}"
    strategy: str = "use_latest_filename"
    compression: str | None = None
    exclude: list = field(default_factory=list)
    ddl: bool = True
    sequences: bool = True
    truncate: bool = False
    data: bool = True
    follow_foreign_keys: bool = False
    strict: bool = False

    @classmethod
    def from_stack(cls, stack: ConfigStack, root_location: str | None = None):
        ddl = stack.get("ddl", True)
        if isinstance(ddl, dict):
            ddl = ddl["enabled"]

        location = compose_root_location(root_location, stack.get("location"), default="backups/{table}")

        return from_partial(
            cls,
            name=stack.get("name"),
            location=location,
            strategy=stack.get("strategy"),
            truncate=stack.get("truncate"),
            compression=stack.get("compression"),
            exclude=stack.get("exclude"),
            sequences=bool(stack.get("sequences", True)),
            data=bool(stack.get("data", True)),
            ddl=bool(ddl),
            follow_foreign_keys=bool(stack.get("follow_foreign_keys", False)),
            strict=bool(stack.get("strict", False)),
        )


@dataclass
class RestoreConfig(TableParentConfig[RestoreTableConfig]):
    @classmethod
    def get_child_class(cls):
        return RestoreTableConfig


def normalize_table_config(tables_config: list | dict) -> list:
    """Convert the dict-style table declaration into list style.

    from: {"name": {"location": "backups/{table}"}}
    to: [{"name": "blah", "location": "backups/{table}"}]

    alternatively:
    from: ["name1", "name2"]
    to: [{"name": "name1"}, {"name": "name2"}]
    """
    if isinstance(tables_config, dict):
        return [{"name": k, **v} for k, v in tables_config.items()]

    if isinstance(tables_config, list):
        return [{"name": c} if isinstance(c, str) else c for c in tables_config]

    return tables_config


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Optional Integration configs which do NOT have default values built-in
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #


@dataclass
class S3Config(Config):
    aws_access_key_id: str | None = None
    aws_secret_access_key: str | None = None
    region: str | None = None
    profile: str | None = None

    @classmethod
    def from_dict(cls, s3_config: dict | None) -> S3Config | None:
        if s3_config is None:
            return None
        return cls(
            aws_access_key_id=s3_config.get("aws_access_key_id"),
            aws_secret_access_key=s3_config.get("aws_secret_access_key"),
            region=s3_config.get("region"),
            profile=s3_config.get("profile"),
        )


def compose_root_location(root_location, location, *, default):
    if root_location is None:
        return location or default

    return join_paths(root_location, location or default)



import os
from typing import Iterable, Optional, Tuple

import click
from sqlalchemy.orm import Session

from databudgie.cli.base import resolver
from databudgie.cli.config import (
    CliConfig,
    collect_config,
    loaders,
    pretty_print,
)
from databudgie.config import (
    BackupConfig,
    ConfigError,
    RestoreConfig,
    RootConfig,
)
from databudgie.manifest.manager import Manifest
from databudgie.output import Console
from databudgie.storage import StorageBackend


@resolver.group()
@click.option("--strict/--no-strict", is_flag=True, default=None)
@click.option("--color/--no-color", is_flag=True, default=True)
@click.option(
    "-a",
    "--adapter",
    default=None,
    type=click.Choice(["postgres", "postgresql", "python"], case_sensitive=False),
    help="Override the automatic dialect detection.",
)
@click.option("-c", "--config", default=None, help="config file", multiple=True)
@click.option(
    "-C",
    "--conn",
    "--connection",
    default=None,
    help="The name of a named connection to use. Takes precedence over the 'url' field if present.",
)
@click.option("-v", "--verbose", count=True, default=0)
@click.option(
    "--ddl/--no-ddl", default=None, is_flag=True, help="Whether to backup the DDL. Overrides the config option, if set"
)
@click.option("-u", "--url", default=None, help="The url used to connect to the database.")
@click.option("-l", "--location", default=None, help="The location to read/write backups from/to.")
@click.option(
    "-t",
    "--table",
    default=None,
    multiple=True,
    help="The set of tables to backup or restore. Note that this overrides any table section in config.",
)
@click.option(
    "-x",
    "--exclude",
    default=None,
    multiple=True,
    help="The set of tables to exclude. Note that this overrides any table section in config.",
)
@click.option(
    "--stats/--no-stats",
    default=None,
    is_flag=True,
    help="Print high level statistics about what the command did. Automatically implied by --dry-run!",
)
@click.option(
    "--dry-run/--no-dry-run",
    default=None,
    is_flag=True,
    help="Do not actually perform the write operations of the backup/restore. It **does**, however, execute the queries.",
)
@click.option(
    "--raw-config",
    default=None,
    help="Accepts raw config as a string, rather than searching a file for it.",
)
@click.option(
    "--raw-config-format",
    default="json",
    help="The assumed format of --raw-config. Defaults to `json`. Must be one of: `yml`, `yaml`, `json`, `toml`.",
    type=click.Choice(loaders),
)
@click.version_option()
def cli(
    strict: bool,
    config: Iterable[str],
    verbose: int = 0,
    color: bool = True,
    conn: Optional[str] = None,
    adapter: Optional[str] = None,
    ddl: Optional[bool] = None,
    url: Optional[str] = None,
    table: Optional[Tuple[str, ...]] = None,
    exclude: Optional[Tuple[str, ...]] = None,
    location: Optional[str] = None,
    dry_run: bool = False,
    stats: bool = False,
    raw_config: Optional[str] = None,
    raw_config_format: str = "json",
):
    if color is False:
        os.environ["NO_COLOR"] = "true"

    if conn and url:
        raise click.UsageError("--url and --connection are mutually exclusive options")

    cli_config = CliConfig(
        ddl=ddl,
        tables=list(table) if table else None,
        exclude=list(exclude) if exclude else None,
        url=url,
        location=location,
        adapter=adapter,
        strict=strict,
    )

    try:
        root_config = collect_config(
            cli_config=cli_config,
            file_names=config,
            raw_config=raw_config,
            raw_config_format=raw_config_format,
        )
    except ConfigError as e:
        raise click.UsageError(*e.args)

    resolver.register_values(
        root_config=root_config,
        verbosity=verbose,
        console=Console(verbosity=verbose),
        connection_name=conn,
        dry_run=bool(dry_run),
        stats=stats if stats is not None else dry_run,
    )


@resolver.command(cli, "backup")
@click.option("--backup-id", default=None, help="Backup manifest id.")
def backup_cli(
    backup_config: BackupConfig,
    backup_db: Session,
    console: Console,
    backup_manifest: Optional[Manifest] = None,
    backup_id: Optional[int] = None,
    stats: bool = False,
    dry_run: bool = False,
):
    """Perform backup."""
    from databudgie.backup import backup_all

    if backup_manifest and backup_id:
        backup_manifest.set_transaction_id(backup_id)

    storage = StorageBackend.from_config(
        backup_config,
        manifest=backup_manifest,
        record_stats=stats,
        perform_writes=not dry_run,
    )

    try:
        backup_all(backup_db, backup_config, storage=storage, console=console)
    except Exception as e:
        console.trace(e)
        raise click.ClickException(str(e))

    if stats:
        storage.print_stats()


@resolver.command(cli, "restore")
@click.option("--restore-id", default=None, help="Restore manifest id.")
@click.option(
    "--clean/--no-clean",
    is_flag=True,
    default=None,
    help="Drops and recreates the target database before performing the restore",
)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    default=False,
    help="Dont ask for confirmation",
)
def restore_cli(
    restore_config: RestoreConfig,
    restore_db: Session,
    console: Console,
    restore_manifest: Optional[Manifest] = None,
    restore_id: Optional[int] = None,
    clean: Optional[bool] = None,
    yes: bool = False,
    stats: bool = False,
    dry_run: bool = False,
):
    """Perform restore."""
    from databudgie.restore import restore_all

    if restore_manifest and restore_id:
        restore_manifest.set_transaction_id(restore_id)

    restore_config.ddl.clean = clean or restore_config.ddl.clean

    if not yes and restore_config.ddl.clean:
        message = "About to delete the database! input 'y' if that's what you want: "
        if input(message) != "y":  # nosec
            return

    if dry_run:
        raise click.UsageError("--dry-run is not (yet) supported for restore")

    storage = StorageBackend.from_config(
        restore_config,
        manifest=restore_manifest,
        record_stats=stats,
        perform_writes=not dry_run,
    )

    try:
        restore_all(restore_db, restore_config=restore_config, storage=storage, console=console)
    except Exception as e:
        console.trace(e)
        raise click.ClickException(*e.args)

    if stats:
        storage.print_stats()


@resolver.command(cli, "config")
def config_cli(root_config: RootConfig):
    """Print dereferenced and populated config."""
    pretty_print(root_config)



from typing import Callable, Optional, Union

import click
import sqlalchemy
import sqlalchemy.engine.url
import sqlalchemy.orm
import strapp.click

from databudgie.config import BackupConfig, RestoreConfig, RootConfig
from databudgie.output import Console

version = getattr(sqlalchemy, "__version__", "")

URLType = Callable[..., sqlalchemy.engine.url.URL]

if version.startswith("1.4") or version.startswith("2."):
    create_url: URLType = sqlalchemy.engine.url.URL.create
else:
    create_url = sqlalchemy.engine.url.URL


def _create_postgres_session(config: Union[BackupConfig, RestoreConfig], connection_name: Optional[str] = None):
    if connection_name:
        connection = config.connections.get(connection_name)
        if connection is None:
            raise click.UsageError(f"Connection '{connection_name}' not found")
    else:
        connection = config.connection
        if connection is None:
            raise click.UsageError("No config found for 'url' field. Either a 'connection' or a 'url' is required")

    url = connection.url
    if isinstance(url, dict):
        url_obj = create_url(**url)
    else:
        url_obj = sqlalchemy.engine.url.make_url(url)

    engine = sqlalchemy.create_engine(url_obj)
    return sqlalchemy.orm.scoping.scoped_session(sqlalchemy.orm.session.sessionmaker(bind=engine))()


def backup_config(root_config: RootConfig):
    if root_config.backup is None:
        raise click.UsageError("No backup config found. Run 'databudgie config' to see your current configuration.")
    return root_config.backup


def restore_config(root_config: RootConfig, console: Console):
    if not root_config.restore:
        raise click.UsageError("No restore config found. Run 'databudgie config' to see your current configuration.")
    return root_config.restore


def backup_db(backup_config: BackupConfig, connection_name: Optional[str] = None):
    return _create_postgres_session(backup_config, connection_name)


def restore_db(restore_config: RestoreConfig, connection_name: Optional[str] = None):
    return _create_postgres_session(restore_config, connection_name)


def backup_manifest(backup_config: BackupConfig, backup_db):
    from databudgie.manifest.manager import BackupManifest

    table_name: Optional[str] = backup_config.manifest
    if table_name:
        return BackupManifest(backup_db, table_name)
    return None


def restore_manifest(restore_config: RestoreConfig, restore_db):
    from databudgie.manifest.manager import RestoreManifest

    table_name: Optional[str] = restore_config.manifest
    if table_name:
        return RestoreManifest(restore_db, table_name)
    return None


resolver = strapp.click.Resolver(
    backup_config=backup_config,
    backup_db=backup_db,
    backup_manifest=backup_manifest,
    restore_config=restore_config,
    restore_db=restore_db,
    restore_manifest=restore_manifest,
)



from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, Sequence, TYPE_CHECKING, TypeVar

from sqlalchemy import inspect
from sqlalchemy.orm import Session

from databudgie.config import BackupTableConfig, RestoreTableConfig
from databudgie.match import expand_table_globs
from databudgie.output import Console, default_console
from databudgie.utils import parse_table

if TYPE_CHECKING:
    from databudgie.storage import StorageBackend

T = TypeVar("T", BackupTableConfig, RestoreTableConfig)


@dataclass
class SchemaOp(Generic[T]):
    name: str
    raw_conf: T

    def location(self) -> str:
        return self.raw_conf.location.format(table=self.name)


@dataclass
class TableOp(Generic[T]):
    """Represents an operation (backup/restore) being performed on a table.

    Fields:
     * `schema`: The expanded name of the table's schema. I.e. the default schema if left unspecified.
     * `table_name`: The raw table name without a schema.
     * `full_name`: The full table name, including the schema.
     * `raw_conf`: directly relates to the raw table config in a backup/restore config.

    Note all `TableOp` records should correspond to concrete tables. That is, given some
    globbed input table "public.*", a `TableOp` will only be produced for a concrete
    table matching that criteria.
    """

    schema: str
    table_name: str
    full_name: str
    raw_conf: T

    @classmethod
    def from_name(cls, full_name: str, raw_conf: T):
        schema, table_name = parse_table(full_name)
        return cls(schema=schema, table_name=table_name, full_name=full_name, raw_conf=raw_conf)

    def location(self) -> str:
        return self.raw_conf.location.format(table=self.full_name)

    def query(self) -> str:
        query = getattr(self.raw_conf, "query")  # RestoreTableConfig has no query attribute
        if query is None:
            query = "SELECT * FROM {table}"

        return query.format(table=self.full_name)

    def schema_op(self) -> SchemaOp:
        return SchemaOp(self.schema, self.raw_conf)


def expand_table_ops(
    session: Session,
    tables: Sequence[T],
    existing_tables: list[str],
    storage: StorageBackend,
    *,
    console: Console = default_console,
    warn_for_unused_tables: bool = False,
) -> list[TableOp[T]]:
    """Produce a full list of table operations to be performed.

    tables in the set of `tables` may be globbed and produce more concrete
    tables than initially specified in the input set.

    Additionally, tables may be filtered, either by the pre-existence of
    manifest data or explicit table exclusions.
    """
    # Avoid hardcoding things like "public", we hardcode this elsewhere, this
    # should probably be moved upstream.
    insp = inspect(session.connection())
    default_schema_name = insp.default_schema_name

    # expand table globs into fully qualified mappings to the config.
    matching_tables: dict[str, list[T]] = {}
    for table_conf in tables:
        pattern = table_conf.name
        if "." not in pattern:
            pattern = f"{default_schema_name}.{pattern}"

        expanded_tables = expand_table_globs(existing_tables, pattern)
        if warn_for_unused_tables and not expanded_tables:
            console.warn(f"Skipping table definition `{pattern}` which did not match any tables.")
            continue

        for table_name in expanded_tables:
            if storage.check_manifest(table_name):
                console.trace(f"Skipping {table_name}...")
                continue

            for exclusion_pattern in table_conf.exclude:
                exclusions = set(expand_table_globs(existing_tables, exclusion_pattern))
                if table_name in exclusions:
                    break

                # Breaking out of this loop implies we've identified that this
                # table should be excluded. Thus backup being gated on unbroken
                # iteration of this loop.
            else:
                matching_tables.setdefault(table_name, []).append(table_conf)

    # Notably, `existing_tables` is assumed to be sorted by table-fk dependencies,
    # which is why this collected separately from this loop, where we iterate
    # over unordered input tables.
    result = []
    for table in existing_tables:
        table_confs = matching_tables.get(table)
        if not table_confs:
            continue

        for table_conf in table_confs:
            table_op = TableOp.from_name(table, raw_conf=table_conf)
            result.append(table_op)

    return result



import gzip
import io
from typing import Dict, Optional, Type


class Compressor:
    name: str = ""
    extension: str = ""

    compressors: Dict[Optional[str], "Compressor"] = {}

    @classmethod
    def get_with_name(cls, name: Optional[str]) -> "Compressor":
        return cls.compressors.get(name, Compressor())

    @classmethod
    def __init_subclass__(cls: Type["Compressor"]):
        Compressor.compressors[cls.name] = cls()

    def compose_filetype(self, filetype: str):
        if not self.extension:
            return filetype

        return ".".join([filetype, self.extension])

    @staticmethod
    def compress(buffer):
        return buffer

    @staticmethod
    def extract(buffer):
        return buffer


class GzipCompressor(Compressor):
    name: str = "gzip"
    extension: str = "gz"

    @staticmethod
    def compress(result):
        buffer = io.BytesIO()
        with gzip.open(buffer, mode="wb") as f:
            f.write(result.getvalue())
        buffer.seek(0)
        return buffer

    @staticmethod
    def extract(buffer):
        return gzip.open(buffer, mode="rb")



import json
from typing import Optional, Sequence, TYPE_CHECKING, Union

import sqlalchemy
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Session

from databudgie.adapter import Adapter
from databudgie.config import RestoreConfig
from databudgie.output import Console, default_console, Progress
from databudgie.storage import FileTypes, StorageBackend
from databudgie.table_op import expand_table_ops, SchemaOp, TableOp
from databudgie.utils import capture_failures, join_paths, parse_table, wrap_buffer

if TYPE_CHECKING:
    pass


def restore_all(
    session: Session,
    restore_config: RestoreConfig,
    storage: Optional[StorageBackend] = None,
    console: Console = default_console,
) -> None:
    """Perform restore on all tables in the config."""
    if storage is None:
        storage = StorageBackend.from_config(restore_config)

    adapter = Adapter.get_adapter(session, restore_config.adapter)

    if restore_config.ddl.clean:
        console.warn("Cleaning database")
        adapter.reset_database()

    restore_all_ddl(
        session,
        restore_config,
        storage=storage,
        adapter=adapter,
        console=console,
    )

    console.trace("Collecting existing tables")
    existing_tables = adapter.collect_existing_tables()

    table_ops = expand_table_ops(
        session,
        restore_config.tables,
        existing_tables,
        storage=storage,
        console=console,
        warn_for_unused_tables=True,
    )

    table_ops = adapter.materialize_table_dependencies(
        table_ops,
        console=console,
        reverse=True,
    )

    restore_sequences(
        session,
        table_ops,
        storage=storage,
        adapter=adapter,
        console=console,
    )
    truncate_tables(
        list(reversed(table_ops)),
        adapter=adapter,
        console=console,
    )
    restore_tables(
        session,
        table_ops,
        storage=storage,
        adapter=adapter,
        console=console,
    )


def restore_all_ddl(
    session: Session,
    restore_config: RestoreConfig,
    *,
    storage: StorageBackend,
    adapter: Adapter,
    console: Console = default_console,
):
    if not restore_config.ddl.enabled:
        return

    ddl_path = restore_config.ddl.location
    strategy = restore_config.ddl.strategy

    with storage.get_file_content(ddl_path, strategy, file_type=FileTypes.manifest) as file_object:
        if not file_object:
            console.info("Found no DDL manifest to restore")
            return

        tables = json.load(file_object.content)

    table_ops = expand_table_ops(
        session, restore_config.tables, storage=storage, existing_tables=tables, console=console
    )
    table_ops = adapter.materialize_table_dependencies(
        table_ops,
        console=console,
        reverse=True,
    )

    schema_names = set()
    schema_ops = []

    for table_op in table_ops:
        schema_op = table_op.schema_op()
        if schema_op.name in schema_names:
            continue

        if not table_op.raw_conf.ddl:
            continue

        schema_names.add(schema_op.name)
        schema_ops.append(schema_op)

    with Progress(console) as progress:
        total = len(schema_ops) + len(table_ops)
        task = progress.add_task("Restoring DDL", total=total)

        for schema_op in schema_ops:
            progress.update(task, description=f"Restoring schema DDL: {schema_op.name}")

            restore_ddl(session, schema_op, ddl_path, storage=storage, console=console)

        for table_op in table_ops:
            progress.update(task, description=f"Restoring DDL: {table_op.full_name}")

            restore_ddl(session, table_op, ddl_path, storage=storage, console=console)

    console.info("Finished Restoring DDL")


def restore_ddl(
    session: Session,
    op: Union[TableOp, SchemaOp],
    ddl_path: str,
    storage: StorageBackend,
    console: Console = default_console,
):
    location = op.location()
    strategy: str = op.raw_conf.strategy

    path = join_paths(ddl_path, location)
    with storage.get_file_content(path, strategy, file_type=FileTypes.ddl) as file_object:
        if not file_object:
            console.warn(f"Found no DDL backups under {path} to restore")
            return

        query = file_object.content.read().decode("utf-8")

    query = "\n".join(
        line
        for line in query.splitlines()
        # XXX: These should be being omitted at the backup stage, it's not the restore process' responsibility!
        if not line.startswith("--")
        and not line.startswith("SET")
        and not line.startswith("SELECT pg_catalog")
        and line
    )
    session.execute(sqlalchemy.text(query))
    session.commit()


def restore_sequences(
    session: Session,
    table_ops: Sequence[TableOp],
    adapter: Adapter,
    storage: StorageBackend,
    console: Console = default_console,
):
    with Progress(console) as progress:
        task = progress.add_task("Restoring sequence positions", total=len(table_ops))

        for table_op in table_ops:
            progress.update(task, description=f"Restoring sequence position: {table_op.full_name}")
            if not table_op.raw_conf.sequences:
                continue

            location = table_op.location()
            strategy: str = table_op.raw_conf.strategy

            path = join_paths(location, "sequences")
            with storage.get_file_content(path, strategy, file_type=FileTypes.sequences) as file_object:
                if not file_object:
                    continue

                sequences = json.load(file_object.content)

            for sequence, value in sequences.items():
                adapter.restore_sequence_value(sequence, value)

    console.info("Finished restoring sequence positions")
    session.commit()


def truncate_tables(table_ops: Sequence[TableOp], adapter: Adapter, console: Console):
    with Progress(console) as progress:
        task = progress.add_task("Truncating Tables", total=len(table_ops))

        for table_op in table_ops:
            data = table_op.raw_conf.data
            truncate = table_op.raw_conf.truncate
            if not data or not truncate:
                continue

            progress.update(task, description=f"[trace]Truncating {table_op.full_name}[/trace]", advance=1)
            adapter.truncate_table(table_op.full_name)

    console.info("Finished truncating tables")


def restore_tables(
    session: Session,
    table_ops: Sequence[TableOp],
    *,
    adapter: Adapter,
    storage: StorageBackend,
    console: Console = default_console,
) -> None:
    with Progress(console) as progress:
        task = progress.add_task("Restoring tables", total=len(table_ops))

        for table_op in table_ops:
            if not table_op.raw_conf.data:
                continue

            progress.update(task, description=f"Restoring table: {table_op.full_name}")

            with capture_failures(strict=table_op.raw_conf.strict):
                restore(
                    session,
                    table_op=table_op,
                    adapter=adapter,
                    storage=storage,
                    console=console,
                )

    console.info("Finished restoring tables")


def restore(
    session: Session,
    *,
    adapter: Adapter,
    storage: StorageBackend,
    table_op: TableOp,
    console: Console = default_console,
) -> None:
    """Restore a CSV file from S3 to the database."""
    # Force table_name to be fully qualified
    schema, table = parse_table(table_op.full_name)
    table_name = f"{schema}.{table}"

    strategy: str = table_op.raw_conf.strategy
    compression = table_op.raw_conf.compression

    with storage.get_file_content(
        table_op.location(),
        strategy,
        file_type=FileTypes.data,
        name=table_op.full_name,
        compression=compression,
    ) as file_object:
        if not file_object:
            console.warn(f"Found no backups for {table_name} to restore")
            return

        with wrap_buffer(file_object.content) as wrapper:
            try:
                adapter.import_csv(wrapper, table_op.full_name)
            except SQLAlchemyError:
                session.rollback()
            else:
                session.commit()

    console.trace(f"Restored {table_name} from {file_object.path}")



import abc
import functools
from typing import Optional

import sqlalchemy
from sqlalchemy import MetaData, Table
from sqlalchemy.orm import Session

from databudgie.utils import parse_table


class Manifest(metaclass=abc.ABCMeta):
    def __init__(self, session: Session, table_name: str, action: str):
        self.table_name = table_name
        self.session = session
        self.action = action

        self._transaction_id: Optional[int] = None

    @functools.lru_cache()
    def manifest_table(self):
        schema, table = parse_table(self.table_name)
        self.metadata = MetaData()
        self.metadata.reflect(bind=self.session.get_bind(), schema=schema, only=[table])
        return Table(table, self.metadata, autoload=True, schema=schema)

    @functools.lru_cache()
    def transaction_id(self):
        if self._transaction_id is None:
            table = self.manifest_table()
            last_transaction = self.session.query(sqlalchemy.func.max(table.c.transaction)).scalar()

            if not last_transaction:
                last_transaction = 0

            return last_transaction + 1

        return self._transaction_id

    def __contains__(self, object) -> bool:
        """Return true if the table_name has an entry for the corresponding transaction id."""
        manifest_table = self.manifest_table()
        return bool(
            self.session.query(manifest_table)
            .filter(manifest_table.c.transaction == self.transaction_id(), manifest_table.c.table == object)
            .first()
        )

    def set_transaction_id(self, id: int):
        self._transaction_id = id

    def record(self, table_name: str, location: str):
        self.session.execute(
            self.manifest_table().insert(),
            [
                {
                    "transaction": self.transaction_id(),
                    "action": self.action,
                    "table": table_name,
                    "file_path": location,
                }
            ],
        )
        self.session.commit()


class BackupManifest(Manifest):
    def __init__(self, session: Session, table_name: str):
        super().__init__(session, table_name, "backup")


class RestoreManifest(Manifest):
    def __init__(self, session: Session, table_name: str):
        super().__init__(session, table_name, "restore")



import fnmatch
from typing import Iterable


def expand_table_globs(existing_tables: Iterable[str], pattern: str):
    """Expand a given pattern into the set of tables which apply to that pattern.

    Examples:
        >>> expand_table_globs(["foo", "bar", "food", "football"], "foo*")
        ['foo', 'food', 'football']
    """
    new_tables = []
    for fq_table_name in existing_tables:
        match = fnmatch.fnmatch(fq_table_name, pattern)
        if not match:
            continue

        new_tables.append(fq_table_name)

    return sorted(new_tables)



from __future__ import annotations

import io
import os
import sys
from dataclasses import asdict, dataclass
from os import isatty
from pathlib import Path
from typing import Iterable, MutableMapping

from configly import Config, JsonLoader, TomlLoader, YamlLoader
from rich.console import Console
from rich.syntax import Syntax
from ruamel.yaml import YAML

from databudgie.config import Config as DatabudgieConfig
from databudgie.config import ConfigStack, RootConfig

DEFAULT_CONFIG_FILES = (
    "config.databudgie.yml",
    "databudgie.yml",
    "databudgie.yaml",
    "databudgie.json",
    "databudgie.toml",
)

loaders: dict[str, type[JsonLoader | TomlLoader | YamlLoader]] = {
    "yaml": YamlLoader,
    "yml": YamlLoader,
    "json": JsonLoader,
    "toml": TomlLoader,
}


@dataclass
class CliConfig(DatabudgieConfig):
    tables: list[str] | None = None
    exclude: list[str] | None = None
    ddl: bool | None = None
    url: str | dict | None = None
    location: str | None = None
    adapter: str | None = None
    strict: bool | None = None

    def to_dict(self) -> dict:
        config = asdict(self)
        return {k: v for k, v in config.items() if v is not None}


def collect_config(
    *,
    cli_config: CliConfig | None = None,
    file_names: Iterable[str] = (),
    raw_config: str | None = None,
    raw_config_format: str = "",
):
    """Collect configuration from various locations into a stacked `RootConfig`.

    Configuration is collected according to the following (descending) priority order:
        - CLI configuration options
        - Environment variable configuration
        - stdin/raw configuration content
        - file-based configuration

    """
    configs = []

    if cli_config:
        configs.append(cli_config.to_dict())

    env_config = collect_env_config()
    configs.append(env_config)

    stdin_config = collect_stdin_config(raw_config_format)
    if stdin_config:
        configs.append(stdin_config)

    if raw_config:
        configs.append(collect_raw_config(content=raw_config, format=raw_config_format))

    if file_names or (not stdin_config and not raw_config):
        # That is, if file-names were provided, always add them. Otherwise, load default files if
        # we're not otherwise receiving configuration through stdin/raw-config.
        configs.extend(load_file_configs(*file_names))

    config_stack = ConfigStack(*configs)
    return RootConfig.from_stack(config_stack)


def collect_env_config(environ: MutableMapping[str, str] = os.environ, prefix="DATABUDGIE_"):
    """Collect environment variables into a config structure."""
    len_prefix = len(prefix)
    env_vars = {k[len_prefix:]: v for k, v in environ.items() if k.startswith(prefix)}

    config: dict = {}
    for env_var, value in env_vars.items():
        *segments, last_segment = env_var.split("__")
        context = config
        for segment in segments:
            context = context.setdefault(segment.lower(), {})

        context[last_segment.lower()] = value

    return config


def collect_stdin_config(format: str) -> dict | None:
    """Collect stdin, if available, as a raw config structure."""
    try:
        is_pipe = not isatty(sys.stdin.fileno())
    except io.UnsupportedOperation:
        is_pipe = False

    if not is_pipe:
        return None

    content = sys.stdin.read()
    return collect_raw_config(content=content, format=format)


def load_file_configs(*file_names: str):
    """Collect configs from the set of input files.

    If no files are provided, falls back to the default file lookup heirarchy.
    """
    stop_when_found = False
    if not file_names:
        stop_when_found = True
        file_names = DEFAULT_CONFIG_FILES

    result = []
    for file_name in file_names:
        path = Path(file_name)
        if not path.exists() and stop_when_found:
            continue

        config = collect_raw_config(file=file_name, format=path.suffix.lower()[1:])
        result.append(config)

        if stop_when_found:
            break

    return result


def collect_raw_config(*, format: str, content: str | None = None, file: str | None = None) -> dict:
    loader_cls = loaders[format]
    config: Config = Config.from_loader(loader_cls(), file=file, content=content)  # type: ignore
    return config.to_dict()


def pretty_print(config: DatabudgieConfig):
    """Pretty print a config model."""
    console = Console()
    buffer = io.StringIO()

    config_as_dict = config.to_dict()
    yaml = YAML()
    yaml.default_flow_style = False
    yaml.dump(config_as_dict, buffer)

    buffer.seek(0)
    data = buffer.read()
    syntax = Syntax(data, "yaml")
    console.print(syntax)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['databudgie', 'databudgie.adapter', 'databudgie.cli', 'databudgie.manifest']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0.0',
 'configly[yaml]',
 'rich',
 'sqlalchemy>=1.3',
 'strapp[click,sqlalchemy]>=0.2.7']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=3.10.0', 'importlib-metadata'],
 'psycopg2': ['psycopg2>=2.7'],
 'psycopg2-binary': ['psycopg2-binary>=2.7'],
 's3': ['boto3']}

entry_points = \
{'console_scripts': ['databudgie = databudgie.__main__:run']}

setup_kwargs = {
    'name': 'databudgie',
    'version': '2.8.0',
    'description': 'Ergonomic and flexible tool for database backup and restore',
    'long_description': '# Databudgie\n\n![Github Actions Build](https://github.com/schireson/databudgie/actions/workflows/build.yml/badge.svg)\n[![Coverage Status](https://coveralls.io/repos/github/schireson/databudgie/badge.svg?branch=main&t=6I0aU6)](https://coveralls.io/github/schireson/databudgie?branch=main)\n[![Documentation\nStatus](https://readthedocs.org/projects/databudgie/badge/?version=latest)](https://databudgie.readthedocs.io)\n\n![](docs/source/_static/databudgie.png)\n\nDatabudgie is a CLI & library for database performing targeted backup and\nrestore of database tables or arbitrary queries against database tables.\n\n# Usage\n\nA minimal config file might look like:\n\n```yaml\n# databudgie.yml or config.databudgie.yml\nbackup:\n  url: postgresql://postgres:postgres@localhost:5432/postgres\n  tables:\n    - name: public.product\n      query: "select * from {table} where store_id > 4"\n      location: s3://my-s3-bucket/databudgie/public.product\nrestore:\n  url: postgresql://postgres:postgres@localhost:5432/postgres\n  tables:\n    - name: public.product\n      location: s3://my-s3-bucket/databudgie/public.product\n```\n\nWith that config in place, backing up the defined tables (using the specified\nconfig) is as simple as `databudgie backup`; and restore `databudgie restore`.\n\n## Installation\n\n```bash\npip install databudgie\n```\n',
    'author': 'Andrew Sosa',
    'author_email': 'andrewso@known.is',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/schireson/databudgie',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)



import contextlib
import io
import os
from typing import Optional, Tuple

from databudgie.output import Console, default_console
from databudgie.s3 import is_s3_path, S3Location


@contextlib.contextmanager
def capture_failures(ignore=(), strict=False, console: Console = default_console):
    """Prevent exceptions from interrupting execution.

    Examples:
        This exception is captured and not propogated:
        >> with capture_failures():
        ...     raise Exception('foo')

        This exception is raised:
        >> with capture_failures(strict=True): # doctest: +IGNORE_EXCEPTION_DETAIL
        ...     raise Exception('foo')
        Traceback (most recent call last):
        Exception: foo

        This exception is also raised:
        >> with capture_failures(ignore=Exception): # doctest: +IGNORE_EXCEPTION_DETAIL
        ...     raise Exception('foo')

    """
    try:
        yield
    except ignore:
        raise
    except Exception as e:
        console.exception(e)
        if strict:
            raise


@contextlib.contextmanager
def wrap_buffer(buffer: io.BytesIO):
    wrapper = io.TextIOWrapper(buffer)
    yield wrapper
    wrapper.detach()
    buffer.seek(0)


def parse_table(table: str) -> Tuple[str, str]:
    """Split a schema-qualified table name into two parts.

    Examples:
        >>> parse_table("myschema.foo")
        ('myschema', 'foo')

        >>> parse_table("bar")
        ('public', 'bar')

        >>> parse_table("...")  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ValueError: Invalid table name: ...
    """
    parts = table.split(".")

    if len(parts) == 1:
        schema = "public"
        table = parts[0]
    elif len(parts) == 2:
        schema, table = parts
    else:
        raise ValueError(f"Invalid table name: {table}")

    return schema, table


def join_paths(*components: Optional[str]) -> str:
    real_components = [c for c in components if c is not None]

    if not real_components:
        return ""

    if len(real_components) == 1:
        return real_components[0]

    first_component, *rest_components = real_components
    normalized_components = []
    for c in rest_components:
        if is_s3_path(c):
            normalized_c = S3Location(c).key
        else:
            normalized_c = c.strip("/")

        normalized_components.append(normalized_c)

    return os.path.join(first_component, *normalized_components)



from rich import console, progress
from rich.table import Table
from rich.theme import Theme
from rich.traceback import Traceback


class Console(console.Console):
    theme = Theme({"trace": "white", "info": "blue", "warn": "yellow", "error": "bold red"})

    def __init__(self, verbosity=0):
        super().__init__(theme=self.theme, log_time=True)
        self.verbosity = verbosity

    def trace(self, message):
        if self.verbosity >= 1:
            return self.log(message, style="trace")
        return None

    def info(self, message):
        return self.log(message, style="info")

    def warn(self, message):
        return self.log(message, style="warn")

    def error(self, message):
        return self.log(message, style="error")

    def exception(self, e):
        tb = Traceback.from_exception(type(e), e, e.__traceback__.tb_next, max_frames=1)
        return self.log(tb, style="error")


class Progress(progress.Progress):
    def __init__(self, console):
        columns = (
            progress.TaskProgressColumn(),
            progress.TimeRemainingColumn(),
            progress.BarColumn(),
            progress.TextColumn("[progress.description]{task.description}"),
        )
        super().__init__(*columns, console=console, transient=True)

    def update(self, task, description, advance=1):
        return super().update(task, description=f"[trace]{description}[/trace]", advance=advance)


default_console = Console()


__all__ = [
    "Console",
    "default_console",
    "Progress",
    "Traceback",
    "Table",
]
